from smc.robots.abstract_robotmanager import AbstractRobotManager
from smc.multiprocessing.process_manager import ProcessManager

from argparse import Namespace
from typing import Any, Callable
import numpy as np
from collections import deque

global control_loop_return
control_loop_return = tuple[bool, dict[str, np.ndarray], dict[str, np.ndarray]]


def PathFollowingFromPlannerControlLoop(
    path_planner: ProcessManager,
    get_position: Callable[[AbstractRobotManager], np.ndarray],
    SOLVER: Any,
    control_loop: Callable[
        [
            Any,
            np.ndarray,
            Namespace,
            AbstractRobotManager,
            int,
            dict[str, deque[np.ndarray]],
        ],
        tuple[np.ndarray, dict[str, np.ndarray], dict[str, np.ndarray]],
    ],
    args: Namespace,
    robot: AbstractRobotManager,
    t: int,  # will be float eventually
    past_data: dict[str, deque[np.ndarray]],
) -> control_loop_return:
    """
    PathFollowingFromPlannerControlLoop
    ---------------
    handles getting path and with comm with the planner.
    then you do whatever you want with the path, and execute a path following controller
    """
    breakFlag = False
    log_item = {}
    save_past_item = {}

    p = get_position(robot)

    path_planner.sendCommand(p)
    data = path_planner.getData()

    if data == "done":
        breakFlag = True

    if data == "done" or data is None:
        robot.sendVelocityCommand(np.zeros(robot.model.nv))
        log_item["qs"] = robot.q
        log_item["dqs"] = robot.v
        return breakFlag, save_past_item, log_item

    _, path2D_untimed = data
    path2D_untimed = np.array(path2D_untimed).reshape((-1, 2))

    v_cmd, past_item_inner, log_item_inner = control_loop(
        SOLVER, path2D_untimed, args, robot, t, past_data
    )
    robot.sendVelocityCommand(v_cmd)
    log_item.update(log_item_inner)
    save_past_item.update(past_item_inner)

    log_item["qs"] = robot.q
    log_item["dqs"] = robot.v
    return breakFlag, save_past_item, log_item
