from smc.robots.abstract_simulated_robotmanager import AbstractSimulatedRobotManager
from smc.robots.interfaces.whole_body_interface import DualArmWholeBodyInterface
from smc.robots.interfaces.mobile_base_interface import (
    get_mobile_base_model,
)
from smc.robots.implementations.yumi import get_yumi_model

import pinocchio as pin
from argparse import Namespace


class AbstractMobileYuMiRobotManager(DualArmWholeBodyInterface):

    def __init__(self, args):
        if args.debug_prints:
            print("MobileYuMiRobotManager init")
        self.args = args
        self._model, self._collision_model, self._visual_model, self._data = (
            get_mobile_yumi_model()
        )

        self._l_ee_frame_id = self.model.getFrameId("robl_tool0")
        self._r_ee_frame_id = self.model.getFrameId("robr_tool0")
        self._base_frame_id = self.model.getFrameId("mobile_base")
        # TODO: CHANGE THIS TO REAL VALUES
        self._MAX_ACCELERATION = 1.7  # const
        self._MAX_QD = 3.14  # const

        self._mode: DualArmWholeBodyInterface.control_mode = (
            DualArmWholeBodyInterface.control_mode.whole_body
        )
        super().__init__(args)


class SimulatedMobileYuMiRobotManager(
    AbstractSimulatedRobotManager, AbstractMobileYuMiRobotManager
):
    def __init__(self, args: Namespace):
        if args.debug_prints:
            print("SimulatedMobileYuMiRobotManager init")
        super().__init__(args)


class RealMobileYumiRobotManager(AbstractMobileYuMiRobotManager):
    def __init__(self, args):
        super().__init__(args)

    # TODO: here assert you need to have ros2 installed to run on real heron
    # and then set all this bullshit here instead of elsewhere
    def set_publisher_joints_cmd(self, publisher_joints_cmd):
        self.publisher_joints_cmd = publisher_joints_cmd
        print("set publisher_joints_cmd into RobotManager")

    def sendVelocityCommand(self, v):
        #        qd_base = qd[:3]
        #        qd_left = qd[3:10]
        #        qd_right = qd[10:]
        #        self.publisher_vel_base(qd_base)
        #        self.publisher_vel_left(qd_left)
        #        self.publisher_vel_right(qd_right)
        empty_msg = JointState()
        for i in range(29):
            empty_msg.velocity.append(0.0)
        msg = empty_msg
        msg.header.stamp = Time().to_msg()
        for i in range(3):
            msg.velocity[i] = qd_cmd[i]
        for i in range(15, 29):
            msg.velocity[i] = qd_cmd[i - 12]

        self.publisher_joints_cmd.publish(msg)


def get_mobile_yumi_model() -> (
    tuple[pin.Model, pin.GeometryModel, pin.GeometryModel, pin.Data]
):

    model_arms, collision_model_arms, visual_model_arms, _ = get_yumi_model()
    model_mobile_base, geom_model_mobile_base = get_mobile_base_model(False)

    # frame-index should be 1
    model, visual_model = pin.appendModel(
        model_mobile_base,
        model_arms,
        geom_model_mobile_base,
        visual_model_arms,
        1,
        pin.SE3.Identity(),
    )
    data = model.createData()

    return model, visual_model.copy(), visual_model, data
