import numpy as np
import pinocchio as pin
from smc.robots.abstract_robotmanager import AbstractRobotManager


class SingleArmInterface(AbstractRobotManager):
    def __init__(self, args):
        if args.debug_prints:
            print("SingleArmInterface init")
        self._ee_frame_id: int
        self._T_w_e: pin.SE3
        super().__init__(args)

    @property
    def ee_frame_id(self):
        return self._ee_frame_id

    @property
    def T_w_e(self):
        return self._T_w_e.copy()

    def computeT_w_e(self, q) -> pin.SE3:
        assert type(q) is np.ndarray
        pin.forwardKinematics(
            self.model,
            self.data,
            q,
        )
        # alternative if you want all frames
        # pin.updateFramePlacements(self.model, self.data)
        pin.updateFramePlacement(self.model, self.data, self._ee_frame_id)
        return self.data.oMf[self._ee_frame_id].copy()

    def getJacobian(self) -> np.ndarray:
        return pin.computeFrameJacobian(
            self.model, self.data, self._q, self._ee_frame_id
        )

    def forwardKinematics(self):
        pin.forwardKinematics(
            self.model,
            self.data,
            self._q,
        )
        pin.updateFramePlacement(self.model, self.data, self._ee_frame_id)
        self._T_w_e = self.data.oMf[self._ee_frame_id].copy()

    def _step(self):
        self._updateQ()
        self._updateV()
        self.forwardKinematics()
